# frozen_string_literal: true

require "dependabot/dependency"
require "dependabot/file_parsers"
require "dependabot/file_parsers/base"
require "dependabot/shared_helpers"
require "dependabot/gradle/version"

# The best Gradle documentation is at:
# - https://docs.gradle.org/current/dsl/org.gradle.api.artifacts.dsl.
#   DependencyHandler.html
#
# In addition, documentation on plugins is at:
# - https://docs.gradle.org/current/userguide/plugins.html
module Dependabot
  module Gradle
    class FileParser < Dependabot::FileParsers::Base
      require "dependabot/file_parsers/base/dependency_set"
      require_relative "file_parser/property_value_finder"

      SUPPORTED_BUILD_FILE_NAMES = %w(build.gradle build.gradle.kts).freeze

      PROPERTY_REGEX =
        /
          (?:\$\{property\((?<property_name>[^:\s]*?)\)\})|
          (?:\$\{(?<property_name>[^:\s]*?)\})|
          (?:\$(?<property_name>[^:\s"']*))
        /x.freeze

      PART = %r{[^\s,@'":/\\]+}.freeze
      VSN_PART = %r{[^\s,'":/\\]+}.freeze
      DEPENDENCY_DECLARATION_REGEX =
        /(?:\(|\s)\s*['"](?<declaration>#{PART}:#{PART}:#{VSN_PART})['"]/.
        freeze
      DEPENDENCY_SET_DECLARATION_REGEX =
        /(?:^|\s)dependencySet\((?<arguments>[^\)]+)\)\s*\{/.freeze
      DEPENDENCY_SET_ENTRY_REGEX = /entry\s+['"](?<name>#{PART})['"]/.freeze
      PLUGIN_BLOCK_DECLARATION_REGEX = /(?:^|\s)plugins\s*\{/.freeze
      PLUGIN_ID_REGEX = /['"](?<id>#{PART})['"]/.freeze

      def parse
        dependency_set = DependencySet.new
        buildfiles.each do |buildfile|
          dependency_set += buildfile_dependencies(buildfile)
        end
        script_plugin_files.each do |plugin_file|
          dependency_set += buildfile_dependencies(plugin_file)
        end
        dependency_set.dependencies
      end

      private

      def map_value_regex(key)
        /(?:^|\s|,|\()#{Regexp.quote(key)}(\s*=|:)\s*['"](?<value>[^'"]+)['"]/
      end

      def buildfile_dependencies(buildfile)
        dependency_set = DependencySet.new

        dependency_set += shortform_buildfile_dependencies(buildfile)
        dependency_set += keyword_arg_buildfile_dependencies(buildfile)
        dependency_set += dependency_set_dependencies(buildfile)
        dependency_set += plugin_dependencies(buildfile)

        dependency_set
      end

      def shortform_buildfile_dependencies(buildfile)
        dependency_set = DependencySet.new

        prepared_content(buildfile).scan(DEPENDENCY_DECLARATION_REGEX) do
          declaration = Regexp.last_match.named_captures.fetch("declaration")

          group, name, version = declaration.split(":")
          version, _packaging_type = version.split("@")
          details = { group: group, name: name, version: version }

          dep = dependency_from(details_hash: details, buildfile: buildfile)
          dependency_set << dep if dep
        end

        dependency_set
      end

      def keyword_arg_buildfile_dependencies(buildfile)
        dependency_set = DependencySet.new

        prepared_content(buildfile).lines.each do |line|
          name    = argument_from_string(line, "name")
          group   = argument_from_string(line, "group")
          version = argument_from_string(line, "version")
          next unless name && group && version

          details = { name: name, group: group, version: version }

          dep = dependency_from(details_hash: details, buildfile: buildfile)
          dependency_set << dep if dep
        end

        dependency_set
      end

      def dependency_set_dependencies(buildfile)
        dependency_set = DependencySet.new

        dependency_set_blocks = []

        prepared_content(buildfile).scan(DEPENDENCY_SET_DECLARATION_REGEX) do
          mch = Regexp.last_match
          dependency_set_blocks <<
            {
              arguments: mch.named_captures.fetch("arguments"),
              block: mch.post_match[0..closing_bracket_index(mch.post_match)]
            }
        end

        dependency_set_blocks.each do |blk|
          group   = argument_from_string(blk[:arguments], "group")
          version = argument_from_string(blk[:arguments], "version")

          next unless group && version

          blk[:block].scan(DEPENDENCY_SET_ENTRY_REGEX).flatten.each do |name|
            dep = dependency_from(
              details_hash: { group: group, name: name, version: version },
              buildfile: buildfile,
              in_dependency_set: true
            )
            dependency_set << dep if dep
          end
        end

        dependency_set
      end

      def plugin_dependencies(buildfile)
        dependency_set = DependencySet.new

        plugin_blocks = []

        prepared_content(buildfile).scan(PLUGIN_BLOCK_DECLARATION_REGEX) do
          mch = Regexp.last_match
          plugin_blocks <<
            mch.post_match[0..closing_bracket_index(mch.post_match)]
        end

        plugin_blocks.each do |blk|
          blk.lines.each do |line|
            name_regex = /(id|kotlin)(\s+#{PLUGIN_ID_REGEX}|\(#{PLUGIN_ID_REGEX}\))/
            name = line.match(name_regex)&.named_captures&.fetch("id")
            version_regex = /version\s+['"](?<version>#{VSN_PART})['"]/
            version = line.match(version_regex)&.named_captures&.
                fetch("version")
            next unless name && version

            details = { name: name, group: "plugins", extra_groups: extra_groups(line), version: version }
            dep = dependency_from(details_hash: details, buildfile: buildfile)
            dependency_set << dep if dep
          end
        end

        dependency_set
      end

      def extra_groups(line)
        line.match(/kotlin(\s+#{PLUGIN_ID_REGEX}|\(#{PLUGIN_ID_REGEX}\))/) ? ["kotlin"] : []
      end

      def argument_from_string(string, arg_name)
        string.
          match(map_value_regex(arg_name))&.
          named_captures&.
          fetch("value")
      end

      def dependency_from(details_hash:, buildfile:, in_dependency_set: false)
        group   = evaluated_value(details_hash[:group], buildfile)
        name    = evaluated_value(details_hash[:name], buildfile)
        version = evaluated_value(details_hash[:version], buildfile)
        extra_groups = details_hash[:extra_groups] || []

        dependency_name =
          if group == "plugins" then name
          else "#{group}:#{name}"
          end
        groups =
          if group == "plugins" then ["plugins"] + extra_groups
          else []
          end
        source =
          source_from(group, name, version)

        # If we can't evaluate a property they we won't be able to
        # update this dependency
        return if "#{dependency_name}:#{version}".match?(PROPERTY_REGEX)
        return unless Gradle::Version.correct?(version)

        Dependency.new(
          name: dependency_name,
          version: version,
          requirements: [{
            requirement: version,
            file: buildfile.name,
            source: source,
            groups: groups,
            metadata: dependency_metadata(details_hash, in_dependency_set)
          }],
          package_manager: "gradle"
        )
      end

      def source_from(group, name, version)
        return nil unless group&.start_with?("com.github") && version.match?(/\A[0-9a-f]{40}\Z/)

        account = group.sub("com.github.", "")

        {
          type: "git",
          url: "https://github.com/#{account}/#{name}",
          branch: nil,
          ref: version
        }
      end

      def dependency_metadata(details_hash, in_dependency_set)
        version_property_name =
          details_hash[:version].
          match(PROPERTY_REGEX)&.
          named_captures&.fetch("property_name")

        return unless version_property_name || in_dependency_set

        metadata = {}
        metadata[:property_name] = version_property_name if version_property_name
        if in_dependency_set
          metadata[:dependency_set] = {
            group: details_hash[:group],
            version: details_hash[:version]
          }
        end
        metadata
      end

      def evaluated_value(value, buildfile)
        return value unless value.scan(PROPERTY_REGEX).count == 1

        property_name  = value.match(PROPERTY_REGEX).
                         named_captures.fetch("property_name")
        property_value = property_value_finder.property_value(
          property_name: property_name,
          callsite_buildfile: buildfile
        )

        return value unless property_value

        value.gsub(PROPERTY_REGEX, property_value)
      end

      def property_value_finder
        @property_value_finder ||=
          PropertyValueFinder.new(dependency_files: dependency_files)
      end

      def prepared_content(buildfile)
        # Remove any comments
        prepared_content =
          buildfile.content.
          gsub(%r{(?<=^|\s)//.*$}, "\n").
          gsub(%r{(?<=^|\s)/\*.*?\*/}m, "")

        # Remove the dependencyVerification section added by Gradle Witness
        # (TODO: Support updating this in the FileUpdater)
        prepared_content.dup.scan(/dependencyVerification\s*{/) do
          mtch = Regexp.last_match
          block = mtch.post_match[0..closing_bracket_index(mtch.post_match)]
          prepared_content.gsub!(block, "")
        end

        prepared_content
      end

      def closing_bracket_index(string)
        closes_required = 1

        string.chars.each_with_index do |char, index|
          closes_required += 1 if char == "{"
          closes_required -= 1 if char == "}"
          return index if closes_required.zero?
        end

        0
      end

      def buildfiles
        @buildfiles ||= dependency_files.select do |f|
          f.name.end_with?(*SUPPORTED_BUILD_FILE_NAMES)
        end
      end

      def script_plugin_files
        @script_plugin_files ||=
          buildfiles.flat_map do |buildfile|
            buildfile.content.
              scan(/apply from(\s+=|:)\s+['"]([^'"]+)['"]/).flatten.
              map { |f| dependency_files.find { |bf| bf.name == f } }.
              compact
          end.
          uniq
      end

      def check_required_files
        raise "No build.gradle or build.gradle.kts!" unless original_file
      end

      def original_file
        dependency_files.find do |f|
          SUPPORTED_BUILD_FILE_NAMES.include?(f.name)
        end
      end
    end
  end
end

Dependabot::FileParsers.register("gradle", Dependabot::Gradle::FileParser)
